#include "gmp.h"

#if defined(_IOSTREAM_) || defined(__IOSTREAM__)
#define _GMPCPP_H_HAVE_IOSTREAM 1
#endif

namespace GMP {

class MPZ;
class MPQ;
class MPF;

class InvalidNumberException {};
class NonexistentInverseException {};
class IntegerInputException {};

// correct some inconsistencies in GMP function names (makes macros simpler)
#define mpf_div_2exp_ui    mpf_div_2exp
#define mpz_fdiv_q_2exp_ui mpz_fdiv_q_2exp
#define mpf_mul_2exp_ui    mpf_mul_2exp
#define mpz_mul_2exp_ui    mpz_mul_2exp
#define mpz_pow_uiui       mpz_ui_pow_ui
#define mpz_root_ui        mpz_root
#define mpz_tdiv_q_2exp_ui mpz_tdiv_q_2exp

template<typename R, typename T, void E(R &, T)> class unary_op {
public:
  unary_op(T op): Op(op) {}
  void eval(R &result) const { E(result, Op); }
private:
  T Op;
};

template<typename R, typename T1, typename T2, void E(R &, T1, T2)> class binary_op {
public:
  binary_op(T1 op1, T2 op2): Op1(op1), Op2(op2) {}
  void eval(R &result) const { E(result, Op1, Op2); }
private:
  T1 Op1;
  T2 Op2;
};

template<typename R, typename T1, typename T2, typename T3, void E(R &, T1, T2, T3)> class ternary_op {
public:
  ternary_op(T1 op1, T2 op2, T3 op3): Op1(op1), Op2(op2), Op3(op3) {}
  void eval(R &result) const { E(result, Op1, Op2, Op3); }
private:
  T1 Op1;
  T2 Op2;
  T3 Op3;
};

#define DECLARE_UNARY_CONSTRUCT_ASSIGN(R, T, evalfunc)             \
  R(const unary_op<R, T, evalfunc> &op)                            \
    { _init(); op.eval(*this); }                                   \
  R &operator =(const unary_op<R, T, evalfunc> &op)                \
    { op.eval(*this); return *this; }

#define DECLARE_UNARY_FUNCTION_STATIC_ONLY(opname, R, T, evalfunc) \
  DECLARE_UNARY_CONSTRUCT_ASSIGN(R, T, evalfunc)                   \
  static unary_op<R, T, evalfunc> opname(T op)                     \
    { return unary_op<R, T, evalfunc>(op); }

#define DECLARE_UNARY_FUNCTION(opname, R, T, evalfunc)             \
  DECLARE_UNARY_FUNCTION_STATIC_ONLY(opname, R, T, evalfunc)       \
  R &opname()                                                      \
    { evalfunc(*this, *this); return *this; }

#define DECLARE_UNARY_OPERATOR(opname, opop, R, T, evalfunc)       \
  DECLARE_UNARY_FUNCTION(opname, R, T, evalfunc)                   \
  unary_op<R, T, evalfunc> opop() const                            \
    { return opname(*this); }

#define DECLARE_BINARY_CONSTRUCT_ASSIGN(R, T1, T2, evalfunc)       \
  R(const binary_op<R, T1, T2, evalfunc> &op)                      \
    { _init(); op.eval(*this); }                                   \
  R &operator =(const binary_op<R, T1, T2, evalfunc> &op)          \
    { op.eval(*this); return *this; }

#define DECLARE_BINARY_FUNCTION_STATIC_ONLY(opname, R, T1, T2, evalfunc) \
  DECLARE_BINARY_CONSTRUCT_ASSIGN(R, T1, T2, evalfunc)             \
  static binary_op<R, T1, T2, evalfunc> opname(T1 op1, T2 op2)     \
    { return binary_op<R, T1, T2, evalfunc>(op1, op2); }

#define DECLARE_BINARY_FUNCTION(opname, R, T1, T2, evalfunc)       \
  DECLARE_BINARY_FUNCTION_STATIC_ONLY(opname, R, T1, T2, evalfunc) \
  R &opname(T2 op)                                                 \
    { evalfunc(*this, *this, op); return *this; }

#define DECLARE_BINARY_OPERATOR(opname, opop, R, T1, T2, evalfunc) \
  DECLARE_BINARY_FUNCTION(opname, R, T1, T2, evalfunc)             \
  binary_op<R, T1, T2, evalfunc> opop(T2 op) const                 \
    { return opname(*this, op); }                                  \
  R &opop##=(T2 op)                                                \
    { evalfunc(*this, *this, op); return *this; }

#define DECLARE_TERNARY_CONSTRUCT_ASSIGN(R, T1, T2, T3, evalfunc)  \
  R(const ternary_op<R, T1, T2, T3, evalfunc> &op)                 \
    { _init(); op.eval(*this); }                                   \
  R &operator =(const ternary_op<R, T1, T2, T3, evalfunc> &op)     \
    { op.eval(*this); return *this; }

#define DECLARE_TERNARY_FUNCTION_STATIC_ONLY(opname, R, T1, T2, T3, evalfunc) \
  DECLARE_TERNARY_CONSTRUCT_ASSIGN(R, T1, T2, T3, evalfunc)        \
  static ternary_op<R, T1, T2, T3, evalfunc> opname(T1 op1, T2 op2, T3 op3) \
    { return ternary_op<R, T1, T2, T3, evalfunc>(op1, op2, op3); }

#define DECLARE_TERNARY_FUNCTION(opname, R, T1, T2, T3, evalfunc)  \
  DECLARE_TERNARY_FUNCTION_STATIC_ONLY(opname, R, T1, T2, T3, evalfunc) \
  R &opname(T2 op2, T3 op3)                                        \
    { evalfunc(*this, *this, op2, op3); return *this; }

#define DECLARE_UNARY_FUNCTION_MPZ(opname)                         \
  static void eval_##opname##_MPZ(MPZ &result, const MPZ &op)      \
    { mpz_##opname(result.rep, op.rep); }                          \
  DECLARE_UNARY_FUNCTION(opname, MPZ, const MPZ &, MPZ::eval_##opname##_MPZ)

#define DECLARE_UNARY_FUNCTION_ulong(opname)                       \
  static void eval_##opname##_ulong(MPZ &result, unsigned long int op) \
    { mpz_##opname##_ui(result.rep, op); }                         \
  DECLARE_UNARY_FUNCTION_STATIC_ONLY(opname, MPZ, unsigned long int, MPZ::eval_##opname##_ulong)

#define DECLARE_UNARY_OPERATOR_MPZ(opname, opop)                   \
  static void eval_##opname##_MPZ(MPZ &result, const MPZ &op)      \
    { mpz_##opname(result.rep, op.rep); }                          \
  DECLARE_UNARY_OPERATOR(opname, opop, MPZ, const MPZ &, MPZ::eval_##opname##_MPZ)

#define DECLARE_BINARY_FUNCTION_MPZ_MPZ(opname)                    \
  static void eval_##opname##_MPZ_MPZ(MPZ &result, const MPZ &op1, const MPZ &op2) \
    { mpz_##opname(result.rep, op1.rep, op2.rep); }                \
  DECLARE_BINARY_FUNCTION(opname, MPZ, const MPZ &, const MPZ &, MPZ::eval_##opname##_MPZ_MPZ)

#define DECLARE_BINARY_FUNCTION_MPZ_ulong(opname)                  \
  static void eval_##opname##_MPZ_ulong(MPZ &result, const MPZ &op1, unsigned long int op2) \
    { mpz_##opname##_ui(result.rep, op1.rep, op2); }               \
  DECLARE_BINARY_FUNCTION(opname, MPZ, const MPZ &, unsigned long int, MPZ::eval_##opname##_MPZ_ulong)

#define DECLARE_BINARY_FUNCTION_ulong_ulong(opname)                \
  static void eval_##opname##_ulong_ulong(MPZ &result, unsigned long int op1, unsigned long int op2) \
    { mpz_##opname##_uiui(result.rep, op1, op2); }                 \
  DECLARE_BINARY_FUNCTION_STATIC_ONLY(opname, MPZ, unsigned long int, unsigned long int, MPZ::eval_##opname##_ulong_ulong)

#define DECLARE_TERNARY_FUNCTION_MPZ_MPZ_MPZ(opname)               \
  static void eval_##opname##_MPZ_MPZ_MPZ(MPZ &result, const MPZ &op1, const MPZ &op2, const MPZ &op3) \
    { mpz_##opname(result.rep, op1.rep, op2.rep, op3.rep); }       \
  DECLARE_TERNARY_FUNCTION(opname, MPZ, const MPZ &, const MPZ &, const MPZ &, MPZ::eval_##opname##_MPZ_MPZ_MPZ)

#define DECLARE_TERNARY_FUNCTION_MPZ_ulong_MPZ(opname)             \
  static void eval_##opname##_MPZ_ulong_MPZ(MPZ &result, const MPZ &op1, unsigned long int op2, const MPZ &op3) \
    { mpz_##opname##_ui(result.rep, op1.rep, op2, op3.rep); }      \
  DECLARE_TERNARY_FUNCTION(opname, MPZ, const MPZ &, unsigned long int, const MPZ &, MPZ::eval_##opname##_MPZ_ulong_MPZ)

#define DECLARE_BINARY_OPERATOR_MPZ_MPZ(opname, opop)              \
  static void eval_##opname##_MPZ_MPZ(MPZ &result, const MPZ &op1, const MPZ &op2) \
    { mpz_##opname(result.rep, op1.rep, op2.rep); }                \
  DECLARE_BINARY_OPERATOR(opname, opop, MPZ, const MPZ &, const MPZ &, MPZ::eval_##opname##_MPZ_MPZ)

#define DECLARE_BINARY_OPERATOR_MPZ_ulong(opname, opop)            \
  static void eval_##opname##_MPZ_ulong(MPZ &result, const MPZ &op1, unsigned long int op2) \
    { mpz_##opname##_ui(result.rep, op1.rep, op2); }               \
  DECLARE_BINARY_OPERATOR(opname, opop, MPZ, const MPZ &, unsigned long int, MPZ::eval_##opname##_MPZ_ulong)

#define DECLARE_BINARY_OPERATOR_MPZ_uint(opname, opop)            \
  static void eval_##opname##_MPZ_uint(MPZ &result, const MPZ &op1, unsigned int op2) \
    { eval_##opname##_MPZ_ulong(result, op1, static_cast<unsigned long int>(op2)); } \
  DECLARE_BINARY_OPERATOR(opname, opop, MPZ, const MPZ &, unsigned int, MPZ::eval_##opname##_MPZ_uint)

#define DECLARE_BINARY_OPERATOR_MPZ_slong(opname, opop)            \
  static void eval_##opname##_MPZ_slong(MPZ &result, const MPZ &op1, signed long int op2) \
    { mpz_##opname##_si(result.rep, op1.rep, op2); }               \
  DECLARE_BINARY_OPERATOR(opname, opop, MPZ, const MPZ &, signed long int, MPZ::eval_##opname##_MPZ_slong)

#define DECLARE_BINARY_OPERATOR_MPZ_sint(opname, opop)            \
  static void eval_##opname##_MPZ_sint(MPZ &result, const MPZ &op1, signed int op2) \
    { eval_##opname##_MPZ_slong(result, op1, static_cast<signed long int>(op2)); } \
  DECLARE_BINARY_OPERATOR(opname, opop, MPZ, const MPZ &, signed int, MPZ::eval_##opname##_MPZ_sint)

#define DECLARE_GLOBAL_BINARY_OPERATOR_ulong_MPZ_co(opname, opop) \
  inline binary_op<MPZ, const MPZ &, unsigned long int, MPZ::eval_##opname##_MPZ_ulong> opop(unsigned long int op1, const MPZ &op2) \
    { return binary_op<MPZ, const MPZ &, unsigned long int, MPZ::eval_##opname##_MPZ_ulong>(op2, op1); }

#define DECLARE_GLOBAL_BINARY_OPERATOR_ulong_MPZ_nc(opname, opop) \
  inline binary_op<MPZ, const MPZ &, const MPZ &, MPZ::eval_##opname##_MPZ_MPZ> opop(unsigned long int op1, const MPZ &op2) \
    { return binary_op<MPZ, const MPZ &, const MPZ &, MPZ::eval_##opname##_MPZ_MPZ>(op1, op2); }

class MPZ {
  void _init() { mpz_init(rep); }
public:
  // Initialization Functions
  MPZ();
  ~MPZ();

  // Compatibility Functions
  operator const __mpz_struct *() const;

  // Assignment Functions
  MPZ &set(const MPZ &op);
  MPZ &set(unsigned long int op);
  MPZ &set(unsigned int op);
  MPZ &set(signed long int op);
  MPZ &set(signed int op);
  MPZ &set(double op);
  MPZ &set(const MPQ &op);
  MPZ &set(const MPF &op);
  MPZ &set(const char *str, int base = 10) throw(InvalidNumberException);
  MPZ &operator =(const MPZ &op);
  MPZ &operator =(unsigned long int op);
  MPZ &operator =(unsigned int op);
  MPZ &operator =(signed long int op);
  MPZ &operator =(signed int op);
  MPZ &operator =(double op);
  MPZ &operator =(const MPQ &op);
  MPZ &operator =(const MPF &op);
  MPZ &operator =(const char *str) throw(InvalidNumberException);
  MPZ &swap(MPZ &rop2);

  // Combined Initialization and Assignment Functions
  MPZ(const MPZ &op);
  MPZ(unsigned long int op);
  MPZ(unsigned int op);
  MPZ(signed long int op);
  MPZ(signed int op);
  MPZ(double op);
  MPZ(const char *str, int base = 10) throw(InvalidNumberException);

  // Conversion Functions
  mp_limb_t getlimbn(mp_size_t n) const;
  unsigned long int get_ui() const;
  signed long int get_si() const;
  double get_d() const;
  char *get_str(char *str = 0, int base = 10) const;

  // Arithmetic Functions

  DECLARE_BINARY_OPERATOR_MPZ_MPZ(add, operator +)
  DECLARE_BINARY_OPERATOR_MPZ_ulong(add, operator +)
  MPZ &operator ++();
  MPZ operator ++(int);
  DECLARE_BINARY_OPERATOR_MPZ_MPZ(sub, operator -)
  DECLARE_BINARY_OPERATOR_MPZ_ulong(sub, operator -)
  MPZ &operator --();
  MPZ operator --(int);
  DECLARE_BINARY_OPERATOR_MPZ_MPZ(mul, operator *)
  DECLARE_BINARY_OPERATOR_MPZ_ulong(mul, operator *)
  DECLARE_BINARY_OPERATOR_MPZ_uint(mul, operator *)
  DECLARE_BINARY_OPERATOR_MPZ_slong(mul, operator *)
  DECLARE_BINARY_OPERATOR_MPZ_sint(mul, operator *)
  MPZ &addmul(const MPZ &op1, unsigned long int op2);
  DECLARE_BINARY_OPERATOR_MPZ_ulong(mul_2exp, operator <<)
  DECLARE_UNARY_OPERATOR_MPZ(neg, operator -)
  DECLARE_UNARY_FUNCTION_MPZ(abs)

  // Division Functions
  DECLARE_BINARY_OPERATOR_MPZ_MPZ(tdiv_q, operator /)
  DECLARE_BINARY_FUNCTION_MPZ_MPZ(tdiv_r)
  unsigned long int tdiv(unsigned long int d) const;
  DECLARE_BINARY_FUNCTION_MPZ_MPZ(fdiv_q)
  DECLARE_BINARY_FUNCTION_MPZ_MPZ(fdiv_r)
  unsigned long int fdiv(unsigned long int d) const;
  DECLARE_BINARY_FUNCTION_MPZ_MPZ(cdiv_q)
  DECLARE_BINARY_FUNCTION_MPZ_MPZ(cdiv_r)
  unsigned long int cdiv(unsigned long int d) const;
  DECLARE_BINARY_OPERATOR_MPZ_MPZ(mod, operator %)
  DECLARE_BINARY_FUNCTION_MPZ_MPZ(divexact)
  DECLARE_BINARY_OPERATOR_MPZ_ulong(tdiv_q_2exp, operator >>)
  DECLARE_BINARY_FUNCTION_MPZ_ulong(fdiv_q_2exp)

  // Exponentiation Functions
  DECLARE_TERNARY_FUNCTION_MPZ_MPZ_MPZ(powm)
  DECLARE_TERNARY_FUNCTION_MPZ_ulong_MPZ(powm)
  DECLARE_BINARY_FUNCTION_MPZ_ulong(pow)
  DECLARE_BINARY_FUNCTION_ulong_ulong(pow)

  // Root Extraction Functions
  DECLARE_BINARY_FUNCTION_MPZ_ulong(root)
  DECLARE_UNARY_FUNCTION_MPZ(sqrt)
  bool perfect_power_p() const;
  bool perfect_square_p() const;

  // Number Theoretic Functions
  int probab_prime_p(int reps) const;
  DECLARE_UNARY_FUNCTION_MPZ(nextprime)
  DECLARE_BINARY_FUNCTION_MPZ_MPZ(gcd)
  static unsigned long int gcd(const MPZ &op1, unsigned long int op2);
  unsigned long int gcd(unsigned long int op) const;
  DECLARE_BINARY_FUNCTION_MPZ_MPZ(lcm)
  static MPZ invert(const MPZ &op1, const MPZ &op2) throw(NonexistentInverseException);
  MPZ &invert(const MPZ &op) throw(NonexistentInverseException);
  static int jacobi(const MPZ &op1, const MPZ &op2);
  int jacobi(const MPZ &op) const;
  static int legendre(const MPZ &op1, const MPZ &op2);
  int legendre(const MPZ &op) const;
  static int kronecker(signed long int a, const MPZ &b);
  static int kronecker(signed int a, const MPZ &b);
  static int kronecker(unsigned long int a, const MPZ &b);
  static int kronecker(unsigned int a, const MPZ &b);
  static int kronecker(const MPZ &a, signed long int b);
  static int kronecker(const MPZ &a, signed int b);
  static int kronecker(const MPZ &a, unsigned long int b);
  static int kronecker(const MPZ &a, unsigned int b);
  DECLARE_BINARY_FUNCTION_MPZ_MPZ(remove)
  DECLARE_UNARY_FUNCTION_ulong(fac)
  DECLARE_BINARY_FUNCTION_MPZ_ulong(bin)
  DECLARE_BINARY_FUNCTION_ulong_ulong(bin)
  DECLARE_UNARY_FUNCTION_ulong(fib)

  // Comparison Functions
  static int cmp(const MPZ &op1, const MPZ &op2);
  static int cmp(const MPZ &op1, unsigned long int op2);
  static int cmp(const MPZ &op1, signed long int op2);
  int cmp(const MPZ &op) const;
  int cmp(unsigned long int op) const;
  int cmp(signed long int op) const;
  bool operator ==(const MPZ &op) const;
  bool operator ==(unsigned long int op) const;
  bool operator ==(unsigned int op) const;
  bool operator ==(signed long int op) const;
  bool operator ==(signed int op) const;
  bool operator !=(const MPZ &op) const;
  bool operator !=(unsigned long int op) const;
  bool operator !=(unsigned int op) const;
  bool operator !=(signed long int op) const;
  bool operator !=(signed int op) const;
  bool operator >(const MPZ &op) const;
  bool operator >(unsigned long int op) const;
  bool operator >(unsigned int op) const;
  bool operator >(signed long int op) const;
  bool operator >(signed int op) const;
  bool operator >=(const MPZ &op) const;
  bool operator >=(unsigned long int op) const;
  bool operator >=(unsigned int op) const;
  bool operator >=(signed long int op) const;
  bool operator >=(signed int op) const;
  bool operator <=(const MPZ &op) const;
  bool operator <=(unsigned long int op) const;
  bool operator <=(unsigned int op) const;
  bool operator <=(signed long int op) const;
  bool operator <=(signed int op) const;
  bool operator <(const MPZ &op) const;
  bool operator <(unsigned long int op) const;
  bool operator <(unsigned int op) const;
  bool operator <(signed long int op) const;
  bool operator <(signed int op) const;
  static int cmpabs(const MPZ &op1, const MPZ &op2);
  static int cmpabs(const MPZ &op1, unsigned long int op2);
  int cmpabs(const MPZ &op) const;
  int cmpabs(unsigned long int op) const;
  int sgn() const;

  // Logical and Bit Manipulation Functions
  DECLARE_BINARY_OPERATOR_MPZ_MPZ(and, operator &)
  DECLARE_BINARY_OPERATOR_MPZ_MPZ(ior, operator |)
  DECLARE_BINARY_OPERATOR_MPZ_MPZ(xor, operator ^)
  DECLARE_UNARY_OPERATOR_MPZ(com, operator ~)
  int popcount() const;
  int hamdist(const MPZ &op) const;
  int scan0(unsigned long int starting_bit) const;
  int scan1(unsigned long int starting_bit) const;
  MPZ &setbit(unsigned long int index);
  MPZ &clrbit(unsigned long int index);
  int tstbit(unsigned long int index) const;

  // Input and Output Functions
#ifdef _GMP_H_HAVE_FILE
  size_t out_str(FILE *stream = NULL, int base = 10) const;
  static MPZ inp_str(FILE *stream = NULL, int base = 10) throw(IntegerInputException);
  size_t out_raw(FILE *stream = NULL) const;
  static MPZ inp_raw(FILE *stream = NULL) throw(IntegerInputException);
#endif

  // Miscellaneous Functions
  bool fits_ulong_p() const;
  bool fits_slong_p() const;
  bool fits_uint_p() const;
  bool fits_sint_p() const;
  bool fits_ushort_p() const;
  bool fits_sshort_p() const;
  bool odd() const;
  bool even() const;
  size_t size() const;
  size_t sizeinbase(int base = 10) const;

  mpz_t rep;
};

DECLARE_GLOBAL_BINARY_OPERATOR_ulong_MPZ_co(add, operator +)
DECLARE_GLOBAL_BINARY_OPERATOR_ulong_MPZ_nc(sub, operator -)
DECLARE_GLOBAL_BINARY_OPERATOR_ulong_MPZ_co(mul, operator *)
DECLARE_GLOBAL_BINARY_OPERATOR_ulong_MPZ_nc(tdiv_q, operator /)
DECLARE_GLOBAL_BINARY_OPERATOR_ulong_MPZ_nc(mod, operator %)

#ifdef _GMPCPP_H_HAVE_IOSTREAM
std::ostream &operator <<(std::ostream &s, const MPZ &op);
std::istream &operator >>(std::istream &s, MPZ &op);
#endif

class RANDSTATE {
public:
  RANDSTATE(gmp_randalg_t randalg = GMP_RAND_ALG_DEFAULT, unsigned long int size = 64);
  RANDSTATE(const MPZ &a, unsigned long int c, unsigned long int m2exp);
  ~RANDSTATE();

  void randseed(const MPZ &seed);
  void randseed(unsigned long int seed);

  MPZ urandomb(unsigned long int n);
  MPZ urandomm(const MPZ &n);
  MPZ rrandomb(unsigned long int n);

  gmp_randstate_t state;

private:
  // copying objects not permitted
  RANDSTATE(const RANDSTATE &);
  RANDSTATE &operator =(const RANDSTATE &);
};

#define DECLARE_UNARY_FUNCTION_MPQ(opname)                         \
  static void eval_##opname##_MPQ(MPQ &result, const MPQ &op)      \
    { mpq_##opname(result.rep, op.rep); }                          \
  DECLARE_UNARY_FUNCTION(opname, MPQ, const MPQ &, MPQ::eval_##opname##_MPQ)

#define DECLARE_UNARY_OPERATOR_MPQ(opname, opop)                   \
  static void eval_##opname##_MPQ(MPQ &result, const MPQ &op)      \
    { mpq_##opname(result.rep, op.rep); }                          \
  DECLARE_UNARY_OPERATOR(opname, opop, MPQ, const MPQ &, MPQ::eval_##opname##_MPQ)

#define DECLARE_BINARY_FUNCTION_MPQ_MPQ(opname)                    \
  static void eval_##opname##_MPQ_MPQ(MPQ &result, const MPQ &op1, const MPQ &op2) \
    { mpq_##opname(result.rep, op1.rep, op2.rep); }                \
  DECLARE_BINARY_FUNCTION(opname, MPQ, const MPQ &, const MPQ &, MPQ::eval_##opname##_MPQ_MPQ)

#define DECLARE_BINARY_OPERATOR_MPQ_MPQ(opname, opop)              \
  static void eval_##opname##_MPQ_MPQ(MPQ &result, const MPQ &op1, const MPQ &op2) \
    { mpq_##opname(result.rep, op1.rep, op2.rep); }                \
  DECLARE_BINARY_OPERATOR(opname, opop, MPQ, const MPQ &, const MPQ &, MPQ::eval_##opname##_MPQ_MPQ)

class MPQ {
  void _init() { mpq_init(rep); }
public:
  // Initialization and Assignment Functions
  MPQ();
  MPQ(const MPQ &op);
  MPQ(unsigned long int op1, unsigned long int op2);
  MPQ(signed long int op1, unsigned long int op2);
  ~MPQ();

  // Compatibility Functions
  operator const __mpq_struct *() const;

  // Assignment Functions
  MPQ &set(const MPQ &op);
  MPQ &set(const MPZ &op);
  MPQ &set(unsigned long int op1, unsigned long int op2);
  MPQ &set(signed long int op1, unsigned long int op2);
  MPQ &operator =(const MPQ &op);
  MPQ &operator =(const MPZ &op);
  MPQ &swap(MPQ &rop2);

  // Arithmetic Functions
  DECLARE_BINARY_OPERATOR_MPQ_MPQ(add, operator +)
  DECLARE_BINARY_OPERATOR_MPQ_MPQ(sub, operator -)
  DECLARE_BINARY_OPERATOR_MPQ_MPQ(mul, operator *)
  DECLARE_BINARY_OPERATOR_MPQ_MPQ(div, operator /)
  DECLARE_UNARY_OPERATOR_MPQ(neg, operator -)
  DECLARE_UNARY_FUNCTION_MPQ(inv)

  // Comparison Functions
  static int cmp(const MPQ &op1, const MPQ &op2);
  static int cmp(const MPQ &op1, unsigned long int num2, unsigned long int den2);
  int cmp(const MPQ &op) const;
  int cmp(unsigned long int num, unsigned long int den) const;
  static int equal(const MPQ &op1, const MPQ &op2);
  int equal(const MPQ &op) const;
  bool operator ==(const MPQ &op) const;
  bool operator !=(const MPQ &op) const;
  bool operator >(const MPQ &op) const;
  bool operator >=(const MPQ &op) const;
  bool operator <=(const MPQ &op) const;
  bool operator <(const MPQ &op) const;
  int sgn() const;

  // Applying Integer Functions to Rationals
  // TODO: these really need to return some sort of MPZref object
  MPZ num();
  MPZ den();

  // Input and Output Functions
#ifdef _GMP_H_HAVE_FILE
  size_t out_str(FILE *stream = NULL, int base = 10) const;
#endif

  // Miscellaneous Functions
  double get_d() const;
  void set_d(double d);
  void set_num(const MPZ &numerator);
  void set_den(const MPZ &denominator);
  MPZ get_num() const;
  MPZ get_den() const;

  mpq_t rep;
};

#ifdef _GMPCPP_H_HAVE_IOSTREAM
std::ostream &operator <<(std::ostream &s, const MPQ &op);
std::istream &operator >>(std::istream &s, MPQ &op);
#endif

#define DECLARE_UNARY_FUNCTION_MPF(opname)                         \
  static void eval_##opname##_MPF(MPF &result, const MPF &op)      \
    { mpf_##opname(result.rep, op.rep); }                          \
  DECLARE_UNARY_FUNCTION(opname, MPF, const MPF &, MPF::eval_##opname##_MPF)

#define DECLARE_UNARY_OPERATOR_MPF(opname, opop)                   \
  static void eval_##opname##_MPF(MPF &result, const MPF &op)      \
    { mpf_##opname(result.rep, op.rep); }                          \
  DECLARE_UNARY_OPERATOR(opname, opop, MPF, const MPF &, MPF::eval_##opname##_MPF)

#define DECLARE_BINARY_FUNCTION_MPF_MPF(opname)                    \
  static void eval_##opname##_MPF_MPF(MPF &result, const MPF &op1, const MPF &op2) \
    { mpf_##opname(result.rep, op1.rep, op2.rep); }                \
  DECLARE_BINARY_FUNCTION(opname, MPF, const MPF &, const MPF &, MPF::eval_##opname##_MPF_MPF)

#define DECLARE_BINARY_FUNCTION_MPF_ulong(opname)                  \
  static void eval_##opname##_MPF_ulong(MPF &result, const MPF &op1, unsigned long int op2) \
    { mpf_##opname##_ui(result.rep, op1.rep, op2); }               \
  DECLARE_BINARY_FUNCTION(opname, MPF, const MPF &, unsigned long int, MPF::eval_##opname##_MPF_ulong)

#define DECLARE_BINARY_FUNCTION_ulong_MPF(opname)                  \
  static void eval_##opname##_ulong_MPF(MPF &result, unsigned long int op1, const MPF &op2) \
    { mpf_ui_##opname(result.rep, op1, op2.rep); }                 \
  DECLARE_BINARY_FUNCTION_STATIC_ONLY(opname, MPF, unsigned long int, const MPF &, MPF::eval_##opname##_ulong_MPF)

#define DECLARE_BINARY_OPERATOR_MPF_MPF(opname, opop)              \
  static void eval_##opname##_MPF_MPF(MPF &result, const MPF &op1, const MPF &op2) \
    { mpf_##opname(result.rep, op1.rep, op2.rep); }                \
  DECLARE_BINARY_OPERATOR(opname, opop, MPF, const MPF &, const MPF &, MPF::eval_##opname##_MPF_MPF)

#define DECLARE_BINARY_OPERATOR_MPF_ulong(opname, opop)            \
  static void eval_##opname##_MPF_ulong(MPF &result, const MPF &op1, unsigned long int op2) \
    { mpf_##opname##_ui(result.rep, op1.rep, op2); }               \
  DECLARE_BINARY_OPERATOR(opname, opop, MPF, const MPF &, unsigned long int, MPF::eval_##opname##_MPF_ulong)

#define DECLARE_BINARY_OPERATOR_MPF_uint(opname, opop)            \
  static void eval_##opname##_MPF_uint(MPF &result, const MPF &op1, unsigned int op2) \
    { eval_##opname##_MPF_ulong(result, op1, static_cast<unsigned long int>(op2)); } \
  DECLARE_BINARY_OPERATOR(opname, opop, MPF, const MPF &, unsigned int, MPF::eval_##opname##_MPF_uint)

#define DECLARE_BINARY_OPERATOR_MPF_slong(opname, opop)            \
  static void eval_##opname##_MPF_slong(MPF &result, const MPF &op1, signed long int op2) \
    { mpf_##opname##_si(result.rep, op1.rep, op2); }               \
  DECLARE_BINARY_OPERATOR(opname, opop, MPF, const MPF &, signed long int, MPF::eval_##opname##_MPF_slong)

#define DECLARE_BINARY_OPERATOR_MPF_sint(opname, opop)            \
  static void eval_##opname##_MPF_sint(MPF &result, const MPF &op1, signed int op2) \
    { eval_##opname##_MPF_slong(result, op1, static_cast<signed long int>(op2)); } \
  DECLARE_BINARY_OPERATOR(opname, opop, MPF, const MPF &, signed int, MPF::eval_##opname##_MPF_sint)

#define DECLARE_GLOBAL_BINARY_OPERATOR_ulong_MPF_co(opname, opop) \
  inline binary_op<MPF, const MPF &, unsigned long int, MPF::eval_##opname##_MPF_ulong> opop(unsigned long int op1, const MPF &op2) \
    { return binary_op<MPF, const MPF &, unsigned long int, MPF::eval_##opname##_MPF_ulong>(op2, op1); }

#define DECLARE_GLOBAL_BINARY_OPERATOR_ulong_MPF_nc(opname, opop) \
  inline binary_op<MPF, const MPF &, const MPF &, MPF::eval_##opname##_MPF_MPF> opop(unsigned long int op1, const MPF &op2) \
    { return binary_op<MPF, const MPF &, const MPF &, MPF::eval_##opname##_MPF_MPF>(op1, op2); }

class MPF {
  void _init() { mpf_init(rep); }
public:
  // Initialization Functions
  static void set_default_prec(unsigned long int prec);
  MPF();
  MPF(bool, unsigned long int prec);
  ~MPF();
  void set_prec(unsigned long int prec);
  unsigned long int get_prec() const;
  void set_prec_raw(unsigned long int prec);

  // Compatibility Functions
  operator const __mpf_struct *() const;

  // Assignment Functions
  MPF &set(const MPF &op);
  MPF &set(unsigned long int op);
  MPF &set(unsigned int op);
  MPF &set(signed long int op);
  MPF &set(signed int op);
  MPF &set(double op);
  MPF &set(const MPZ &op);
  MPF &set(const MPQ &op);
  MPF &set(const char *str, int base = 10) throw(InvalidNumberException);
  MPF &operator =(const MPF &op);
  MPF &operator =(unsigned long int op);
  MPF &operator =(unsigned int op);
  MPF &operator =(signed long int op);
  MPF &operator =(signed int op);
  MPF &operator =(double op);
  MPF &operator =(const MPZ &op);
  MPF &operator =(const MPQ &op);
  MPF &operator =(const char *str) throw(InvalidNumberException);
  MPF &swap(MPF &rop2);

  // Combined Initialization and Assignment Functions
  MPF(const MPF &op);
  MPF(unsigned long int op);
  MPF(unsigned int op);
  MPF(signed long int op);
  MPF(signed int op);
  MPF(double op);
  MPF(const char *str, int base = 10) throw(InvalidNumberException);

  // Conversion Functions
  double get_d() const;
  char *get_str(char *str = 0, mp_exp_t *expptr = 0, int base = 10, size_t ndigits = 0) const;

  // Arithmetic Functions
  DECLARE_BINARY_OPERATOR_MPF_MPF(add, operator +)
  DECLARE_BINARY_OPERATOR_MPF_ulong(add, operator +)
  MPF &operator ++();
  MPF operator ++(int);
  DECLARE_BINARY_OPERATOR_MPF_MPF(sub, operator -)
  DECLARE_BINARY_OPERATOR_MPF_ulong(sub, operator -)
  DECLARE_BINARY_FUNCTION_ulong_MPF(sub)
  MPF &operator --();
  MPF operator --(int);
  DECLARE_BINARY_OPERATOR_MPF_MPF(mul, operator *)
  DECLARE_BINARY_OPERATOR_MPF_ulong(mul, operator *)
  DECLARE_BINARY_OPERATOR_MPF_MPF(div, operator /)
  DECLARE_BINARY_OPERATOR_MPF_ulong(div, operator /)
  DECLARE_BINARY_FUNCTION_ulong_MPF(div)
  DECLARE_UNARY_FUNCTION_MPF(sqrt)
  //DECLARE_UNARY_FUNCTION_ulong(sqrt)
  DECLARE_BINARY_FUNCTION_MPF_ulong(pow)
  DECLARE_UNARY_OPERATOR_MPF(neg, operator -)
  DECLARE_UNARY_FUNCTION_MPF(abs)
  DECLARE_BINARY_OPERATOR_MPF_ulong(mul_2exp, operator <<)
  DECLARE_BINARY_OPERATOR_MPF_ulong(div_2exp, operator >>)

  // Comparison Functions
  static int cmp(const MPF &op1, const MPF &op2);
  static int cmp(const MPF &op1, unsigned long int op2);
  static int cmp(const MPF &op1, signed long int op2);
  int cmp(const MPF &op) const;
  int cmp(unsigned long int op) const;
  int cmp(signed long int op) const;
  bool operator ==(const MPF &op) const;
  bool operator ==(unsigned long int op) const;
  bool operator ==(unsigned int op) const;
  bool operator ==(signed long int op) const;
  bool operator ==(signed int op) const;
  bool operator !=(const MPF &op) const;
  bool operator !=(unsigned long int op) const;
  bool operator !=(unsigned int op) const;
  bool operator !=(signed long int op) const;
  bool operator !=(signed int op) const;
  bool operator >(const MPF &op) const;
  bool operator >(unsigned long int op) const;
  bool operator >(unsigned int op) const;
  bool operator >(signed long int op) const;
  bool operator >(signed int op) const;
  bool operator >=(const MPF &op) const;
  bool operator >=(unsigned long int op) const;
  bool operator >=(unsigned int op) const;
  bool operator >=(signed long int op) const;
  bool operator >=(signed int op) const;
  bool operator <=(const MPF &op) const;
  bool operator <=(unsigned long int op) const;
  bool operator <=(unsigned int op) const;
  bool operator <=(signed long int op) const;
  bool operator <=(signed int op) const;
  bool operator <(const MPF &op) const;
  bool operator <(unsigned long int op) const;
  bool operator <(unsigned int op) const;
  bool operator <(signed long int op) const;
  bool operator <(signed int op) const;
  static int eq(const MPF &op1, const MPF &op2, unsigned long int op3);
  int eq(const MPF &op2, unsigned long int op3) const;
  DECLARE_BINARY_FUNCTION_MPF_MPF(reldiff)
  int sgn() const;

  // Input and Output Functions
#ifdef _GMP_H_HAVE_FILE
  size_t out_str(FILE *stream = NULL, int base = 10, size_t n_digits = 0) const;
  static MPF inp_str(FILE *stream = NULL, int base = 10) throw(IntegerInputException);
#endif

  // Miscellaneous Functions
  DECLARE_UNARY_FUNCTION_MPF(ceil)
  DECLARE_UNARY_FUNCTION_MPF(floor)
  DECLARE_UNARY_FUNCTION_MPF(trunc)

  mpf_t rep;
};

DECLARE_GLOBAL_BINARY_OPERATOR_ulong_MPF_co(add, operator +)
DECLARE_GLOBAL_BINARY_OPERATOR_ulong_MPF_nc(sub, operator -)
DECLARE_GLOBAL_BINARY_OPERATOR_ulong_MPF_co(mul, operator *)
DECLARE_GLOBAL_BINARY_OPERATOR_ulong_MPF_nc(div, operator /)

#ifdef _GMPCPP_H_HAVE_IOSTREAM
std::ostream &operator <<(std::ostream &s, const MPF &op);
std::istream &operator >>(std::istream &s, MPF &op);
#endif

#include "gmpcppi.h"

} // namespace GMP
